if not Doodleboard then Doodleboard = {} end
local revision = tonumber(("$Rev: 34 $"):match("(%d+)")) or 1000;
local libmg = LibStub("LibMouseGestures-1.0");
local AceConfigDialog = LibStub("AceConfigDialog-3.0");
local Doodlepad = {messages = {}};
local colors = {{1,0,0,1},{0,1,0,1},{0,0,1,1},{1,1,0,1},{1,0,1,1},{0,1,1,1},{1,0.5,0,1},{0,0,0,1}}
local bases = { '0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z' }; -- for rebasing
local OpenDoodles = {};
local dw,dh = 480, 320; -- default width/height. Used for the uniform coordinate system.
local penWidth = 80; -- default pen size.
local myName = UnitName("player"); -- my name...or your name!
local syntax = "S1"; -- Syntax version control
local Storyboard_Frame = nil;
Doodleboard.raidRank = Doodleboard.raidRank or 1;
Doodleboard.guildRank = Doodleboard.guildRank or 10;

function rebase(number, from, to)
	local num = tonumber(number, from) or 0;
	local y,retval,rem = to - 1,"",num;
	while ( num > y ) do local x = math.floor(num/to); rem = num - (x*to); num = x;retval = (bases[rem+1] or 0) .. (retval or 0);end
	retval = (bases[num+1] or 0) .. ( retval or 0);
	return retval;
end

Doodlepad.consoleOptions = {
	name = "Doodlepad",
	type = 'group',
	args = {
		confdesc = {
			order = 1,
			type = "description",
			name = "Here you can set the general configuration of Doodlepad.",
			cmdHidden = true
		},
		monitorGroup = {
			order = 5,
			name = "Incoming doodles",
			type = "group",
			inline = true,
			args = {
				raidRank = {
				order = 1,
				name  = "Doodle collaboration in raids",
				desc = "This value sets who will be allowed to collaborate on doodles sent by others in a raid.",
				type = 'select',
				values = {'Everyone','Raid Leader + Assisstant', 'Raid Leader'},
				get = function(info) return Doodleboard.raidRank; end,
				set = function(info,v)
					Doodleboard.raidRank = v;
					end,
				},
				guildRank = {
				order = 1,
				name  = "Doodle collaboration in guilds",
				desc = "This value sets who will be allowed to collaborate on doodles sent by others in the guild.",
				type = 'select',
				values = {'Guild Master','Rank 1','Rank 2', 'Rank 3', 'Rank 4', 'Rank 5', 'Rank 6', 'Rank 7', 'Rank 8', 'Everyone'},
				get = function(info) return Doodleboard.guildRank; end,
				set = function(info,v)
					Doodleboard.guildRank = v;
					end,
				},
			}
		}
	}
}
LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Doodlepad", Doodlepad.consoleOptions)
AceConfigDialog:AddToBlizOptions("Doodlepad", "Doodlepad")


function Doodlepad_window(chan, uid, silent)
	uid = uid or "";
    local frame = CreateFrame("Frame", nil, UIParent);
    frame:SetBackdrop({bgFile = "Interface/GLUES/loading",edgeFile = "Interface/GLUES/Common/TextPanel-Border",edgeSize = 16,insets = { left = 4, right = 4, top = 4, bottom = 4 }});
    frame:SetBackdropColor(1,1,1,0.4);
    frame:SetWidth(dw)
    frame:SetHeight(dh);
    frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
	frame:RegisterForDrag("LeftButton");
	frame:SetScript("OnDragStart",function()  frame:StartMoving(); end);
	frame:SetScript("OnDragStop", function()  frame:StopMovingOrSizing(); end);
	frame:SetMovable(true);
	frame:EnableMouse(true);
	frame:EnableMouseWheel(true);
	frame:SetResizable(true);
	local subFrame = CreateFrame("Frame");
	subFrame:SetParent(frame);
	subFrame:SetPoint("TOPLEFT", 5, -28);
	subFrame:SetPoint("BOTTOMRIGHT", -5, 15);
	local canvas = subFrame:CreateTexture("ARTWORK");
	canvas:SetTexture(1,1,1,0.85);
	canvas:SetAllPoints(subFrame);
	local recorder = libmg:New(subFrame);
	local xframe = frame;
	if ( uid:len() == 16 ) then recorder.uid = uid; frame:Hide();
	else recorder.uid = ("%08x%08x"):format(math.random(time()), math.random(GetTime())); end
	recorder.editbox = CreateFrame("EditBox", "DIALOG", recorder.drawLayer, "InputBoxTemplate");
    recorder.editbox:SetAutoFocus(true);
    recorder.editbox:SetScript("OnEscapePressed", function(me) me:Hide(); end);
    recorder.editbox:SetScript("OnEnterPressed",  function(me)
		local txt = {text=me:GetText(), x = me.x*(dw/recorder.width),y=me.y*(dh/recorder.height)};
		tinsert(recorder.cdoodle, txt);
		Doodle_Draw(recorder);
		tinsert(recorder.doodles, recorder.cdoodle);
		Doodle_Broadcast_Prepare(recorder, recorder.cdoodle);
		recorder.cdoodle = {creator=myName};
		me:Hide();
		end);
    recorder.editbox:SetTextInsets(0, 0, 3, 3);
    recorder.editbox:SetMaxLetters(100);
    recorder.editbox:Hide();

	-- caption
	local caption = frame:CreateFontString(nil, "DIALOG");
	caption:SetParent(frame);
	caption:SetFontObject(ChatFontNormal);
	caption:SetText("Doodlepad");
	caption:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5);
	caption:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 5, -25);
	recorder.caption = caption;

    -- close button
    local close = CreateFrame("button", nil, frame, "UIPanelCloseButton");
    close:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0);
    close:Show();
	close:SetScript("OnClick",
		function()
		frame:Hide();
		Doodle_Draw(recorder, true); -- wipe the pad
		print("Doodlepad closed. The doodle has been saved in your storyboard for later use.");
		Storyboard_Save(recorder);
		Storyboard_Scroll(0);
		end);

	-- resizer
    local sizer = CreateFrame("frame", nil, frame);
    sizer:SetPoint("BOTTOMRIGHT")
    sizer:SetHeight(15);
    sizer:SetWidth(15);
    sizer:EnableMouse()
    sizer:SetScript("OnMouseDown",function(frame) frame:GetParent():StartSizing("BOTTOMRIGHT") end)
    sizer:SetScript("OnMouseUp", function(frame) frame:GetParent():StopMovingOrSizing() recorder.width = xframe:GetWidth(); recorder.height = xframe:GetHeight(); Doodle_Redraw(recorder); end)

    local line1 = sizer:CreateTexture(nil, "BACKGROUND")
    line1:SetWidth(14)
    line1:SetHeight(14)
    line1:SetPoint("BOTTOMRIGHT", 0, 0)
    line1:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
    local x = 0.1 * 14/17
    line1:SetTexCoord(0.05 - x, 0.5, 0.05, 0.5 + x, 0.05, 0.5 - x, 0.5 + x, 0.5)

    local line2 = sizer:CreateTexture(nil, "BACKGROUND")
    line2:SetWidth(8)
    line2:SetHeight(8)
    line2:SetPoint("BOTTOMRIGHT", 0,0)
    line2:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
    local x = 0.1 * 8/17
    line2:SetTexCoord(0.05 - x, 0.5, 0.05, 0.5 + x, 0.05, 0.5 - x, 0.5 + x, 0.5)

	recorder.width, recorder.height = dw,dh;
	recorder.doodles, recorder.cdoodle, recorder.texs, recorder.free, recorder.icon, recorder.color, recorder.strings = {},{creator=myName},{},{}, 1, 1,{};
	recorder.target = recorder.drawLayer:CreateTexture("OVERLAY",7);
	recorder.target:SetTexture([[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_]] .. recorder.icon);
	recorder.target:SetWidth(32);
	recorder.target:SetHeight(32);

	frame:SetScript("OnMouseWheel",
		function(caller, value)
			if ( IsControlKeyDown() ) then
				recorder.icon = recorder.icon + value;
				if ( recorder.icon > 8 ) then recorder.icon = 1; end
				if ( recorder.icon < 1 ) then recorder.icon = 8; end
			else
				recorder.color = recorder.color + value;
				if ( recorder.color > #colors ) then recorder.color = 1; end
				if ( recorder.color < 1 ) then recorder.color = #colors; end
			end
		end);
	subFrame:SetScript("OnUpdate",
		function()
			if ( subFrame:IsMouseOver() ) then
				if ( IsControlKeyDown() ) then
					recorder.target:SetTexture([[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_]] .. recorder.icon);
					recorder.target:SetWidth(32);
					recorder.target:SetHeight(32);
				else
					recorder.target:SetTexture(unpack(colors[recorder.color or 1]));
					recorder.target:SetWidth(4);
					recorder.target:SetHeight(4);
				end
				recorder.target:SetPoint("CENTER", recorder.drawLayer, "TOPLEFT", recorder.mouseEnd[1], -recorder.mouseEnd[2]);
				recorder.target:Show();
			else
				recorder.target:Hide();
			end
		end);
	chan = string.lower(chan or "");
	if ( chan == "raid" or chan == "party" or chan == "guild" or chan == "battleground" ) then
		recorder.chan = chan;
		if not silent then print("Sharing doodle with the " .. chan); caption:SetText("Doodlepad: Broadcasting to the " .. chan); end
	elseif ( chan:len() > 1 ) then
		recorder.chan = chan;
		if not silent then print("Sharing doodle with " .. chan); caption:SetText("Doodlepad: Sharing with " .. chan); end
	end
	local found = false;
	for n = 1, #OpenDoodles do
		if OpenDoodles[n].uid == recorder.uid then found = true break end
	end
	if not found then
		tinsert(OpenDoodles, {uid=recorder.uid, recorder=recorder, sender=myName, painters={myName}, frame=frame});
	end
	recorder:StartCapture(Doodlepad.defaultScript);
	return frame, recorder;
end

function Doodle_DrawLine(rec, T, sx, sy, ex, ey, col)
	sx = sx * (rec.width/dw);
	sy = sy * (rec.height/dh);
	ex = ex * (rec.width/dw);
	ey = ey * (rec.height/dh);
	T:SetTexCoord(0,1,0,1);
	local color = col or color;
    sy = -sy;
    ey = -ey;
	local C = rec.drawLayer;
    w = penWidth;
    local dx,dy = ex - sx, ey - sy;
    local cx,cy = (sx + ex) / 2, (sy + ey) / 2;
    if (dx < 0) then
        dx,dy = -dx,-dy;
    end
    local Z = (256/255) / 2;
    local l = sqrt((dx * dx) + (dy * dy));
    local s,c = -dy / l, dx / l;
    local sc = s * c;
    local Bwid, Bhgt, BLx, BLy, TLx, TLy, TRx, TRy, BRx, BRy;
    if (dy >= 0) then
        Bwid = ((l * c) - (w * s)) * Z;
        Bhgt = ((w * c) - (l * s)) * Z;
        BLx, BLy, BRy = (w / l) * sc, s * s, (l / w) * sc;
        BRx, TLx, TLy, TRx = 1 - BLy, BLy, 1 - BRy, 1 - BLx;
        TRy = BRx;
    else
        Bwid = ((l * c) + (w * s)) * Z;
        Bhgt = ((w * c) + (l * s)) * Z;
        BLx, BLy, BRx = s * s, -(l / w) * sc, 1 + (w / l) * sc;
        BRy, TLx, TLy, TRy = BLx, 1 - BRx, 1 - BLx, 1 - BLy;
        TRx = TLy;
    end
    T:SetDrawLayer("BACKGROUND", 3)
	T:ClearAllPoints();
	if (tonumber(tostringall(Bwid)) == nil or tonumber(tostringall(Bhgt)) == nil) then return; end -- discard bad data (#IND or #INF)
    T:SetVertexColor(color[1],color[2],color[3],color[4]);
    T:SetTexCoord(TLx, TLy, BLx, BLy, TRx, TRy, BRx, BRy);
    T:SetPoint("TOPLEFT",   C, "TOPLEFT", cx - Bwid, cy + Bhgt);
    T:SetWidth(Bwid*2);
    T:SetHeight(Bhgt*2);
    T:Show()
end

function Doodlepad_Incoming(prefix,msg,chan,sender)
	if ( prefix == "DPD2" ) then
		local uid, syn, typ, data = msg:sub(1,16), msg:sub(17,18), msg:sub(19,19), msg:sub(20);
		if syn ~= syntax then return end
		local found = false;
		local rec = nil;
		for n = 1, #OpenDoodles do
			if ( OpenDoodles[n].uid == uid ) then
				found = true;
				rec = OpenDoodles[n].recorder;
				break;
			end
		end
		if ( not found and sender ~= UnitName("player")) then
			local frame, recorder = Doodlepad_window(chan, uid,true);
			recorder.caption:SetText("Doodlepad - Broadcast by " .. sender);
			tinsert(OpenDoodles, { frame = frame, recorder = recorder, uid = uid, sender = sender} );
			rec = recorder;
			rec.sender = sender;
			rec.painters = {sender};
			rec.uid = uid;
			if ( chan == "WHISPER" ) then chan = sender; end
			rec.chan = chan;
			print(string.format("Doodlepad: %s is broadcasting a doodle. |HDPD:%s|h|cFFFFFF00[Click here]|r|h to view it",sender, uid));
		end
		if ( rec ) then
			local allowed = false;
			if ( sender == rec.sender ) then allowed = true; end
			if ( (chan == "RAID" or chan == "BATTLEGROUND") and not allowed ) then
				for n = 1, 25 do
					name, rank, _ = GetRaidRosterInfo(n);
					if ( name == sender ) then
						if ( rank >= Doodleboard.raidRank ) then
							allowed = true;
							local found = false;
							for i = 1, #rec.painters do
								if ( rec.painters[n] == sender ) then found = true; break; end
							end
							if ( not found ) then
								print( ("Doodlepad: %s has joined %s's doodle session"):format(sender, rec.sender));
								tinsert(rec.painters, sender);
							end
							break;
						end
					end
				end
			elseif ( chan == "GUILD" and not allowed ) then
				local total = GetNumGuildMembers(true)
				for i = 1, total do
					local name, rank, rankIndex, level, class, zone, note, officernote, online, status, classFileName = GetGuildRosterInfo(i)
					if ( name == sender ) then
						if ( rankIndex <= Doodleboard.guildRank ) then allowed = true; end
						break;
					end
				end
			end
			if ( (allowed and sender ~= UnitName("player")) or rec.chan == UnitName("player") ) then
				if ( #rec.cdoodle > 0 ) then
					tinsert(rec.doodles, rec.cdoodle);
					rec.cdoodle = {creator=sender};
				end
				if ( typ == "D" ) then -- drawings
					local color, data = data:sub(1,6), data:sub(7);
					local Color = {}; string.gsub(color, "([0-9a-fA-F][0-9a-fA-F])", function (x) tinsert(Color, tonumber(x, 16) / 255); end);
					rec.cdoodle.creator = sender;
					rec.cdoodle.color = Color;
					for n = 1, math.floor(data:len()/5) do
						local p = ((n-1)*5)+1;
						local x, y, z = rebase(data:sub(p, p+1), 32, 10), rebase(data:sub(p+2, p+3), 32, 10), data:sub(p+4, p+4);
						tinsert(rec.cdoodle, {x,y,nil,z});
						Doodle_Draw(rec, false, Color);
					end
					tinsert(rec.doodles, rec.cdoodle);
					rec.cdoodle = {creator=sender};
				elseif ( typ == "U" ) then Doodlepad_Undo(rec, sender);
				elseif ( typ == "W" ) then Doodlepad_Undo(rec, sender, true);
				elseif ( typ == "T" ) then
					local x, y, t = rebase(data:sub(1,2), 32, 10), rebase(data:sub(3,4), 32, 10), data:sub(5);
					tinsert(rec.cdoodle, {text=t, x=x,y=y});
					Doodle_Draw(rec);
					tinsert(rec.doodles, rec.cdoodle);
					rec.cdoodle = {creator=sender};
				end
			end
		end
	end
end

function Doodle_ParseLink(link, text, button, chatframe)
	local linktype, uid = strsplit(":", link)
	if linktype == "DPD" then
		for n = 1, #OpenDoodles do
			if OpenDoodles[n].uid == uid then OpenDoodles[n].frame:Show() break end
		end
	else
		Doodlepad.itemref(link, text, button, chatframe)
	end
end

function Doodle_Update(rec,a,b,c,d)
	if ( #rec.cdoodle > 0 ) then
		local l = rec.cdoodle[#rec.cdoodle];
		if ( math.floor(l[1]) ~= math.floor(c) or math.floor(l[2]) ~= math.floor(d) ) then
			local dist = math.sqrt( math.pow(l[1] - c, 2) + math.pow(l[2] - d, 2));
			if ( dist > 4 ) then -- Any lines shorter than 4 pixels get discarded.
				c,d = c*(dw/rec.width), d*(dh/rec.height);
				tinsert(rec.cdoodle, {c,d,nil,0});
			end
		end
	else
		c,d = c*(dw/rec.width), d*(dh/rec.height);
		tinsert(rec.cdoodle, {c,d,nil,0});
	end
	Doodle_Draw(rec, nil, rec.cdoodle.color);
end

function Doodle_Start(rec,a,b,c,d)
	rec.cdoodle.color = colors[rec.color];
	local now = GetTime();
	if ( rec.lastClick and (now - rec.lastClick) < 0.35) then -- double click for placing text
		rec.editbox:ClearAllPoints();
		rec.editbox:SetPoint("CENTER", rec.drawLayer, "TOPLEFT", c, -d);
		rec.editbox:SetWidth(96);
		rec.editbox:SetHeight(16);
		rec.editbox:SetText("");
		rec.editbox.x,rec.editbox.y = c,d;
		rec.editbox:SetJustifyH("CENTER");
		rec.editbox:SetTextInsets(0, 0, 3, 3);
		rec.editbox:Show();
	end
	rec.lastClick = now;
end

function Doodle_Stop(rec,a,b,c,d)
	if IsControlKeyDown() then -- Insert raid target icon if control key is down
		tinsert(rec.cdoodle, {c*(dw/rec.width),d*(dh/rec.height), nil, rec.icon or 1});
		Doodle_Draw(rec);
	end
	tinsert(rec.doodles, rec.cdoodle);
	rec:StartCapture(Doodlepad.defaultScript);
	Doodle_Broadcast_Prepare(rec, rec.cdoodle);
	rec.cdoodle = {creator=myName};
end

function Doodle_Broadcast_Prepare(rec, doodle)
	if ( rec.chan and rec.chan:len() > 1 and (#doodle > 1 or (doodle[1] and doodle[1][4] ~= 0)) or (doodle[1] and doodle[1].text)) then
		rec.chan = (rec.chan or ""):lower();
		if ( rec.chan:len() < 2 or (rec.uid or ""):len() == 0 ) then return end
		local x = 0;
		local c = "";
		if ( not doodle[1].text ) then
			local color = ("%02x%02x%02x"):format(doodle.color[1]*255, doodle.color[2]*255, doodle.color[3]*255);
			for n = 1, #doodle do
				x = x + 1;
				c = c .. ("%2s%2s%1u"):format(rebase(math.floor(doodle[n][1]), 10, 32), rebase(math.floor(doodle[n][2]), 10, 32), doodle[n][4] or 0);
				if ( x > 40 ) then
					if ( rec.chan == "raid" or rec.chan == "guild" or rec.chan == "party" or rec.chan == "battleground" ) then
						tinsert(Doodlepad.messages, {rec.uid .. syntax .. "D" .. color .. c, rec.chan});
					else
						tinsert(Doodlepad.messages, {rec.uid .. syntax .."D" .. color .. c, "WHISPER", rec.chan});
					end
					c = ("%2s%2s%1u"):format(rebase(math.floor(doodle[n][1]), 10, 32), rebase(math.floor(doodle[n][2]), 10, 32), doodle[n][4] or 0);
					x = 0;
				end
			end
			if ( c:len() > 0 ) then
				if ( rec.chan == "raid" or rec.chan == "guild" or rec.chan == "party" or rec.chan == "battleground" ) then
					tinsert(Doodlepad.messages, {rec.uid .. syntax .."D" .. color .. c, rec.chan});
				else
					tinsert(Doodlepad.messages, {rec.uid .. syntax .."D" .. color .. c, "WHISPER", rec.chan});
				end
			end
		else
			local c = ("%2s%2s"):format(rebase(math.floor(doodle[1].x), 10, 32), rebase(math.floor(doodle[1].y), 10, 32));
			if ( rec.chan == "raid" or rec.chan == "guild" or rec.chan == "party"  or rec.chan == "battleground" ) then
				tinsert(Doodlepad.messages, {rec.uid .. syntax .."T" .. c .. doodle[1].text, rec.chan});
			else
				tinsert(Doodlepad.messages, {rec.uid .. syntax .."T" .. c .. doodle[1].text, "WHISPER", rec.chan});
			end
		end
	end
end

function Doodlepad_Undo(rec, author, wipe)
	if wipe then rec.doodles = {};
	else
		for n = #rec.doodles, 1, -1 do
			local doodle = rec.doodles[#rec.doodles];
			if ( doodle and (doodle.creator == author or not doodle.creator)) then
				tremove(doodle);
				if ( #doodle == 0 ) then
					tremove(rec.doodles, n);
				end
				break;
			end
		end
	end
	Doodle_Redraw(rec);
end

function Doodle_Cancel(rec)
	local ret = "U";
	if IsShiftKeyDown() then -- wipe the canvas if shift is down.
		ret = "W"; Doodlepad_Undo(rec, myName, true);
	else
		Doodlepad_Undo(rec, myName);
	end
	if ( rec.chan and rec.uid and rec.uid:len() == 16) then
		if ( rec.chan == "raid" or rec.chan == "guild" or rec.chan == "party" or rec.chan == "battleground" ) then
			tinsert(Doodlepad.messages, {rec.uid .. syntax ..ret, rec.chan});
		elseif ( rec.chan:len() > 1 ) then
			tinsert(Doodlepad.messages, {rec.uid .. syntax ..ret, "WHISPER", rec.chan});
		end
	end
	rec.cdoodle = {creator=myName};
	rec:StartCapture(Doodlepad.defaultScript);
end

function Doodle_Draw_Point(rec, pointa, pointb, col)
	if ( tonumber(pointa[4]) == 0 ) then
		if ( not pointa[3] and pointb ) then
			local T = tremove(rec.free) or rec.drawLayer:CreateTexture();
			T:SetTexture([[Interface\AddOns\Doodlepad\line]]);
			pointa[3] = T; tinsert(rec.texs, T);
			Doodle_DrawLine(rec, T, pointb[1], pointb[2], pointa[1], pointa[2], col);
		end
	elseif ( tonumber(pointa[4]) ) then
		local x,y = pointa[1], pointa[2];
		x,y = x * (rec.width/dw), y * (rec.height/dh);
		local T = tremove(rec.free) or rec.drawLayer:CreateTexture("OVERLAY", 7);
		T:SetTexture([[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_]] .. pointa[4] or 1);
		T:SetVertexColor(1,1,1,1);
		T:SetWidth(32);T:SetHeight(32);T:SetPoint("TOPLEFT", x-16, -y+16);
		tinsert(rec.texs, T); pointa[3] = T;
		T:Show();
	elseif ( pointa.text ) then
		local x,y = pointa.x, pointa.y;
		x,y = x * (rec.width/dw), y * (rec.height/dh);
		local caption = rec.drawLayer:CreateFontString(nil, "DIALOG");
		caption:SetParent(rec.drawLayer);
		caption:SetFontObject(SystemFont_Shadow_Med1);
		caption:SetTextColor(0.3,0.4,0.9,1);
		caption:SetText(pointa.text);
		caption:SetPoint("CENTER", rec.drawLayer, "TOPLEFT", x, -y);
		tinsert(rec.strings, caption);
	end
end

function Doodle_Draw(rec, wipe, col)
	if ( wipe == true ) then
		for n = 1, #rec.texs do
			local T = tremove(rec.texs);
			T:Hide();
			T:SetTexCoord(0,1,0,1);
			tinsert(rec.free, T);
		end
		for n = 1, #rec.strings do
			local T = tremove(rec.strings);
			T:Hide();
		end
	else
		for n = 1, #rec.cdoodle do Doodle_Draw_Point(rec, rec.cdoodle[n], rec.cdoodle[n-1], col); end
	end
end

function Doodle_Redraw(rec)
	Doodle_Draw(rec, true)
	for n = 1, #rec.doodles do
		local doodle = rec.doodles[n];
		local col = doodle.color or colors[5];
		for n = 1, #doodle do doodle[n][3] = nil; Doodle_Draw_Point(rec, doodle[n], doodle[n-1], col); end
	end
	local doodle = rec.cdoodle;
	local col = doodle.color or colors[5];
	for n = 1, #doodle do doodle[n][3] = nil; Doodle_Draw_Point(rec, doodle[n], doodle[n-1], col); end
end


function Storyboard_Save(rec)
	local doodle = {
		uid = rec.uid or ("%08x%08x"):format(math.random(time()), math.random(GetTime())),
		sender = rec.sender or UnitName("player"),
		doodles = rec.doodles
	};
	local canvas;
	for n = 1, #Doodleboard do
		if ( Doodleboard[n].uid == doodle.uid ) then
			Doodleboard[n].doodles = {};
			canvas = Doodleboard[n].doodles;
			break;
		end
	end
	if ( not canvas ) then
		Doodleboard[#Doodleboard+1] = {uid=doodle.uid, creator = doodle.sender, doodles = {}};
		canvas = Doodleboard[#Doodleboard].doodles;
	end
	for n = 1, #doodle.doodles do
		local doodle = doodle.doodles[n];
		if ( doodle[1].text ) then
			local c = ("T%2s%2s%s"):format(rebase(math.floor(doodle[1].x or 0), 10, 32), rebase(math.floor(doodle[1].y or 0), 10, 32), doodle[1].text or "");
			tinsert(canvas, c);
		else
			if not doodle.color then doodle.color = {0,0,0} end
			local c = ("D%02x%02x%02x"):format(doodle.color[1]*255, doodle.color[2]*255, doodle.color[3]*255);
			for i = 1, #doodle do
				c = c .. ("%02s%02s%1u"):format(rebase(math.floor(doodle[i][1] or 0), 10, 32), rebase(math.floor(doodle[i][2] or 0), 10, 32), doodle[i][4] or 0);
			end
			tinsert(canvas, c);
		end
	end
end

function Storyboard_Load(uid)
	local doodle;
	local retdoodle = {};
	for n = 1, #Doodleboard do
		if ( Doodleboard[n].uid == uid ) then
			doodle = Doodleboard[n];
			break;
		end
	end
	if (doodle) then
		retdoodle.uid = doodle.uid;
		retdoodle.creator = doodle.creator;
		retdoodle.doodles = {};
		retdoodle.cdoodle = {creator=myName};
		for n = 1, #doodle.doodles do
			local d = doodle.doodles[n];
			local t = d:sub(1,1);
			local data = d:sub(2);
			if ( t == "D" ) then
				local color, data = data:sub(1,6), data:sub(7);
				local Color = {}; string.gsub(color, "([0-9a-fA-F][0-9a-fA-F])", function (x) tinsert(Color, tonumber(x, 16) / 255); end);
				local s = {creator=myName, color=Color};
				for n = 1, math.floor(data:len()/5) do
					local p = ((n-1)*5)+1;
					local x, y, z = rebase(data:sub(p, p+1), 32, 10), rebase(data:sub(p+2, p+3), 32, 10), data:sub(p+4, p+4);
					tinsert(s, {x,y,nil,z});
				end
				tinsert(retdoodle.doodles, s);
			elseif (t == "T" ) then
				local x, y, t = rebase(data:sub(1,2), 32, 10), rebase(data:sub(3,4), 32, 10), data:sub(5);
				tinsert(retdoodle.doodles, {{x=x,y=y,text=t}});
			end
		end
	end
	return retdoodle;
end

function Storyboard_Delete(uid)
	for n = 1, #Doodleboard do
		if ( Doodleboard[n].uid == uid ) then
			tremove(Doodleboard, n);
			break;
		end
	end
	Storyboard_Scroll(0);
end

function Storyboard_Scroll(value)
	Storyboard_Frame.X = (Storyboard_Frame.X or 1) - value;
	if ( (Storyboard_Frame.X+2) > #Doodleboard ) then Storyboard_Frame.X = #Doodleboard - 2; end
	if ( Storyboard_Frame.X < 1 ) then Storyboard_Frame.X = 1; end
	local i = 0;
	for n = Storyboard_Frame.X, Storyboard_Frame.X+2 do
		i = i + 1;
		if ( Doodleboard[n] ) then
			local doodle = Storyboard_Load(Doodleboard[n].uid);
			Storyboard_Frame.doodles[i].doodles = doodle.doodles;
			Storyboard_Frame.doodles[i].uid = doodle.uid;
			Storyboard_Frame.doodles[i].creator = doodle.creator;
			Doodle_Redraw(Storyboard_Frame.doodles[i]);
		else
			Storyboard_Frame.doodles[i].doodles = {};
			Storyboard_Frame.doodles[i].uid = nil;
			Storyboard_Frame.doodles[i].creator = myName;
			Doodle_Redraw(Storyboard_Frame.doodles[i]);
		end
	end
end

function Storyboard_Context(rec)
	local menu = {
    { text = "Tools:", isTitle = true},
    { text = "Open doodle", func = function() local f,r = Doodlepad_window(""); r.uid = rec.uid; r.doodles = rec.doodles; Doodle_Redraw(r); f:Show(); f:Raise(); Storyboard_Frame:Hide(); end },
	{ text = "Open and broadcast to:", hasArrow = true, menuList = {
        { text = "Raid", func = function(caller) EasyMenu({}, Storyboard_Frame.menuFrame, "cursor", 0 , 0, "MENU"); local f,r = Doodlepad_window("raid"); r.uid = rec.uid; r.doodles = rec.doodles; Doodle_Redraw(r); f:Show(); f:Raise(); Storyboard_Frame:Hide(); for n=1, #rec.doodles do Doodle_Broadcast_Prepare(r, rec.doodles[n]); end end },
		{ text = "Party", func = function(caller) EasyMenu({}, Storyboard_Frame.menuFrame, "cursor", 0 , 0, "MENU"); local f,r = Doodlepad_window("party"); r.uid = rec.uid; r.doodles = rec.doodles; Doodle_Redraw(r); f:Show(); f:Raise(); Storyboard_Frame:Hide();for n=1, #rec.doodles do Doodle_Broadcast_Prepare(r, rec.doodles[n]); end  end },
		{ text = "Guild", func = function(caller) EasyMenu({}, Storyboard_Frame.menuFrame, "cursor", 0 , 0, "MENU");local f,r = Doodlepad_window("guild"); r.uid = rec.uid; r.doodles = rec.doodles; Doodle_Redraw(r); f:Show(); f:Raise(); Storyboard_Frame:Hide(); for n=1, #rec.doodles do Doodle_Broadcast_Prepare(r, rec.doodles[n]); end end },
		{ text = "Battleground", func = function(caller) EasyMenu({}, Storyboard_Frame.menuFrame, "cursor", 0 , 0, "MENU");local f,r = Doodlepad_window("battleground"); r.uid = rec.uid; r.doodles = rec.doodles; Doodle_Redraw(r); f:Show(); f:Raise(); Storyboard_Frame:Hide(); for n=1, #rec.doodles do Doodle_Broadcast_Prepare(r, rec.doodles[n]); end end },
		}
    },
	{ text = "Open as new doodle", func = function() local f,r = Doodlepad_window(""); r.uid = nil; r.doodles = rec.doodles; Doodle_Redraw(r); f:Show(); f:Raise(); Storyboard_Frame:Hide(); end },
	{ text = "Delete doodle", func = function() Storyboard_Delete(rec.uid) end },
    { text = "Cancel"},
    };
    EasyMenu(menu, Storyboard_Frame.menuFrame, "cursor", 0 , 0, "MENU");
end

---- Storyboard ----
Storyboard_Frame = CreateFrame("Frame", nil, UIParent);
Storyboard_Frame:SetBackdrop({bgFile = "Interface/GLUES/loading",edgeFile = "Interface/GLUES/Common/TextPanel-Border",edgeSize = 16,insets = { left = 4, right = 4, top = 4, bottom = 4 }});
Storyboard_Frame:SetBackdropColor(1,1,1,0.4);
Storyboard_Frame:SetWidth(640)
Storyboard_Frame:SetHeight(220);
Storyboard_Frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
Storyboard_Frame:RegisterForDrag("LeftButton");
Storyboard_Frame:SetScript("OnDragStart",function()  Storyboard_Frame:StartMoving(); end);
Storyboard_Frame:SetScript("OnDragStop", function()  Storyboard_Frame:StopMovingOrSizing(); end);
Storyboard_Frame:SetScript("OnMouseWheel", function(caller, value)  Storyboard_Scroll(value); end);
Storyboard_Frame:SetMovable(true);
Storyboard_Frame:EnableMouse(true);
Storyboard_Frame:EnableMouseWheel(true);
Storyboard_Frame:Hide();
Storyboard_Frame.menuFrame = CreateFrame("Frame", "Storyboard_Frame_Menu", UIParent, "UIDropDownMenuTemplate")

--Container frames
Storyboard_Frame.doodles = {};
local x = 10;
for n = 1, 3 do
	Storyboard_Frame.doodles[n] = CreateFrame("Frame", nil, Storyboard_Frame);
	Storyboard_Frame.doodles[n]:SetWidth(200);Storyboard_Frame.doodles[n]:SetHeight(150);
	Storyboard_Frame.doodles[n].canvas = Storyboard_Frame.doodles[n]:CreateTexture("BACKGROUND", -7);
	Storyboard_Frame.doodles[n].drawLayer = CreateFrame("Frame", nil, Storyboard_Frame.doodles[n])
	Storyboard_Frame.doodles[n].drawLayer:SetAllPoints();
	Storyboard_Frame.doodles[n].canvas:SetTexture(1,1,1,1);
	Storyboard_Frame.doodles[n].canvas:SetAllPoints();
	Storyboard_Frame.doodles[n]:SetPoint("TOPLEFT", x,-40);
	Storyboard_Frame.doodles[n].texs = {};
	Storyboard_Frame.doodles[n].free = {};
	Storyboard_Frame.doodles[n].strings = {};
	Storyboard_Frame.doodles[n].cdoodle = {};
	Storyboard_Frame.doodles[n].doodles = {};
	Storyboard_Frame.doodles[n].width = 200;
	Storyboard_Frame.doodles[n].height = 150;
	Storyboard_Frame.doodles[n].drawLayer:EnableMouse(true);
	Storyboard_Frame.doodles[n].drawLayer:SetScript("OnMouseup", function(caller,event) if ( event == "RightButton" ) then Storyboard_Context(Storyboard_Frame.doodles[n]) end end);
	x = x + 210;
end

-- caption
local caption = Storyboard_Frame:CreateFontString(nil, "DIALOG");
caption:SetParent(Storyboard_Frame);
caption:SetFontObject(ChatFontNormal);
caption:SetText("Storyboard");
caption:SetPoint("TOPLEFT", Storyboard_Frame, "TOPLEFT", 5, -5);
caption:SetPoint("BOTTOMRIGHT", Storyboard_Frame, "TOPRIGHT", 5, -25);

caption = Storyboard_Frame:CreateFontString(nil, "DIALOG");
caption:SetParent(Storyboard_Frame);
caption:SetFontObject(ChatFontNormal);
caption:SetText("Use your mouse wheel to scroll through the saved doodles. Right click a doodle for file actions.");
caption:SetPoint("TOPLEFT", Storyboard_Frame, "BOTTOMLEFT", 5, 25);
caption:SetPoint("BOTTOMRIGHT", Storyboard_Frame, "BOTTOMRIGHT", 5, 5);

-- close button
local close = CreateFrame("button", nil, Storyboard_Frame, "UIPanelCloseButton");
close:SetPoint("TOPRIGHT", Storyboard_Frame, "TOPRIGHT", 0, 0);
close:Show();

function Doodle_Broadcast()
	if ( #Doodlepad.messages > 0 ) then
		local now = GetTime();
		Doodlepad.last = Doodlepad.last or now;
		if ( now - Doodlepad.last >= 0.6 ) then
			Doodlepad.last = now;
			local msg, chan, target = unpack(tremove(Doodlepad.messages, 1));
			SendAddonMessage("DPD2", msg, chan, target);
		end
	end
end

Doodlepad.defaultScript = {
	startButton		= "LeftButtonDown",
	stopButton		= "LeftButtonUp",
	cancelButton    = "RightButtonUp",
	startFunc 		= Doodle_Start,
	updateFunc		= Doodle_Update,
	stopFunc 		= Doodle_Stop,
	cancelFunc		= Doodle_Cancel,
};

_G.SLASH_DOODLE1 = '/doodlepad';
_G.SLASH_STORYBOARD1 = '/storyboard';
SlashCmdList['DOODLE'] = function(chan) Doodlepad_window(chan) end;
SlashCmdList['STORYBOARD'] = function(chan) Storyboard_Scroll(0); Storyboard_Frame:Show(); end;
local frame = CreateFrame("Frame");
frame:SetParent(UIParent);frame:SetPoint("TOPLEFT");
frame:RegisterEvent("CHAT_MSG_ADDON");
frame:SetScript("OnEvent", function(caller, event, ...) if ( event == "CHAT_MSG_ADDON" ) then Doodlepad_Incoming(...) end end);
frame:SetScript("OnUpdate", Doodle_Broadcast);
Doodlepad.itemref = _G.SetItemRef;
_G.SetItemRef = Doodle_ParseLink;
print("Doodlepad |cffff8816[Doodleclysm!]|r r"..revision.." loaded.");
